# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2018  Free Software Foundation Europe e.V.
#
# This file is part of reuse, available from its original location:
# <https://git.fsfe.org/reuse/reuse/>.
#
# reuse is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# reuse is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# reuse.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""A list with all SPDX licenses.

Last updated 2017-12-28.
"""

#: All licenses in the SPDX specification.
LICENSES = [
    '0BSD',
    'AAL',
    'Abstyles',
    'Adobe-2006',
    'Adobe-Glyph',
    'ADSL',
    'AFL-1.1',
    'AFL-1.2',
    'AFL-2.0',
    'AFL-2.1',
    'AFL-3.0',
    'Afmparse',
    'AGPL-1.0',
    'AGPL-3.0-only',
    'AGPL-3.0-or-later',
    'AGPL-3.0',
    'Aladdin',
    'AMDPLPA',
    'AML',
    'AMPAS',
    'ANTLR-PD',
    'Apache-1.0',
    'Apache-1.1',
    'Apache-2.0',
    'APAFML',
    'APL-1.0',
    'APSL-1.0',
    'APSL-1.1',
    'APSL-1.2',
    'APSL-2.0',
    'Artistic-1.0-cl8',
    'Artistic-1.0-Perl',
    'Artistic-1.0',
    'Artistic-2.0',
    'Bahyph',
    'Barr',
    'Beerware',
    'BitTorrent-1.0',
    'BitTorrent-1.1',
    'Borceux',
    'BSD-1-Clause',
    'BSD-2-Clause-FreeBSD',
    'BSD-2-Clause-NetBSD',
    'BSD-2-Clause-Patent',
    'BSD-2-Clause',
    'BSD-3-Clause-Attribution',
    'BSD-3-Clause-Clear',
    'BSD-3-Clause-LBNL',
    'BSD-3-Clause-No-Nuclear-License-2014',
    'BSD-3-Clause-No-Nuclear-License',
    'BSD-3-Clause-No-Nuclear-Warranty',
    'BSD-3-Clause',
    'BSD-4-Clause-UC',
    'BSD-4-Clause',
    'BSD-Protection',
    'BSD-Source-Code',
    'BSL-1.0',
    'bzip2-1.0.5',
    'bzip2-1.0.6',
    'Caldera',
    'CATOSL-1.1',
    'CC0-1.0',
    'CC-BY-1.0',
    'CC-BY-2.0',
    'CC-BY-2.5',
    'CC-BY-3.0',
    'CC-BY-4.0',
    'CC-BY-NC-1.0',
    'CC-BY-NC-2.0',
    'CC-BY-NC-2.5',
    'CC-BY-NC-3.0',
    'CC-BY-NC-4.0',
    'CC-BY-NC-ND-1.0',
    'CC-BY-NC-ND-2.0',
    'CC-BY-NC-ND-2.5',
    'CC-BY-NC-ND-3.0',
    'CC-BY-NC-ND-4.0',
    'CC-BY-NC-SA-1.0',
    'CC-BY-NC-SA-2.0',
    'CC-BY-NC-SA-2.5',
    'CC-BY-NC-SA-3.0',
    'CC-BY-NC-SA-4.0',
    'CC-BY-ND-1.0',
    'CC-BY-ND-2.0',
    'CC-BY-ND-2.5',
    'CC-BY-ND-3.0',
    'CC-BY-ND-4.0',
    'CC-BY-SA-1.0',
    'CC-BY-SA-2.0',
    'CC-BY-SA-2.5',
    'CC-BY-SA-3.0',
    'CC-BY-SA-4.0',
    'CDDL-1.0',
    'CDDL-1.1',
    'CDLA-Permissive-1.0',
    'CDLA-Sharing-1.0',
    'CECILL-1.0',
    'CECILL-1.1',
    'CECILL-2.0',
    'CECILL-2.1',
    'CECILL-B',
    'CECILL-C',
    'ClArtistic',
    'CNRI-Jython',
    'CNRI-Python-GPL-Compatible',
    'CNRI-Python',
    'Condor-1.1',
    'CPAL-1.0',
    'CPL-1.0',
    'CPOL-1.02',
    'Crossword',
    'CrystalStacker',
    'CUA-OPL-1.0',
    'Cube',
    'curl',
    'D-FSL-1.0',
    'diffmark',
    'DOC',
    'Dotseqn',
    'DSDP',
    'dvipdfm',
    'ECL-1.0',
    'ECL-2.0',
    'eCos-2.0',
    'EFL-1.0',
    'EFL-2.0',
    'eGenix',
    'Entessa',
    'EPL-1.0',
    'EPL-2.0',
    'ErlPL-1.1',
    'EUDatagrid',
    'EUPL-1.0',
    'EUPL-1.1',
    'EUPL-1.2',
    'Eurosym',
    'Fair',
    'Frameworx-1.0',
    'FreeImage',
    'FSFAP',
    'FSFULLR',
    'FSFUL',
    'FTL',
    'GFDL-1.1-only',
    'GFDL-1.1-or-later',
    'GFDL-1.1',
    'GFDL-1.2-only',
    'GFDL-1.2-or-later',
    'GFDL-1.2',
    'GFDL-1.3-only',
    'GFDL-1.3-or-later',
    'GFDL-1.3',
    'Giftware',
    'GL2PS',
    'Glide',
    'Glulxe',
    'gnuplot',
    'GPL-1.0-only',
    'GPL-1.0-or-later',
    'GPL-1.0',
    'GPL-1.0+',
    'GPL-2.0-only',
    'GPL-2.0-or-later',
    'GPL-2.0-with-autoconf-exception',
    'GPL-2.0-with-bison-exception',
    'GPL-2.0-with-classpath-exception',
    'GPL-2.0-with-font-exception',
    'GPL-2.0-with-GCC-exception',
    'GPL-2.0',
    'GPL-2.0+',
    'GPL-3.0-only',
    'GPL-3.0-or-later',
    'GPL-3.0-with-autoconf-exception',
    'GPL-3.0-with-GCC-exception',
    'GPL-3.0',
    'GPL-3.0+',
    'gSOAP-1.3b',
    'HaskellReport',
    'HPND',
    'IBM-pibs',
    'ICU',
    'IJG',
    'ImageMagick',
    'iMatix',
    'Imlib2',
    'Info-ZIP',
    'Intel-ACPI',
    'Intel',
    'Interbase-1.0',
    'IPA',
    'IPL-1.0',
    'ISC',
    'JasPer-2.0',
    'JSON',
    'LAL-1.2',
    'LAL-1.3',
    'Latex2e',
    'Leptonica',
    'LGPL-2.0-only',
    'LGPL-2.0-or-later',
    'LGPL-2.0',
    'LGPL-2.0+',
    'LGPL-2.1-only',
    'LGPL-2.1-or-later',
    'LGPL-2.1',
    'LGPL-2.1+',
    'LGPL-3.0-only',
    'LGPL-3.0-or-later',
    'LGPL-3.0',
    'LGPL-3.0+',
    'LGPLLR',
    'Libpng',
    'libtiff',
    'LiLiQ-P-1.1',
    'LiLiQ-R-1.1',
    'LiLiQ-Rplus-1.1',
    'LPL-1.02',
    'LPL-1.0',
    'LPPL-1.0',
    'LPPL-1.1',
    'LPPL-1.2',
    'LPPL-1.3a',
    'LPPL-1.3c',
    'MakeIndex',
    'MirOS',
    'MIT-advertising',
    'MIT-CMU',
    'MIT-enna',
    'MIT-feh',
    'MITNFA',
    'MIT',
    'Motosoto',
    'mpich2',
    'MPL-1.0',
    'MPL-1.1',
    'MPL-2.0-no-copyleft-exception',
    'MPL-2.0',
    'MS-PL',
    'MS-RL',
    'MTLL',
    'Multics',
    'Mup',
    'NASA-1.3',
    'Naumen',
    'NBPL-1.0',
    'NCSA',
    'NetCDF',
    'Net-SNMP',
    'Newsletr',
    'NGPL',
    'NLOD-1.0',
    'NLPL',
    'Nokia',
    'NOSL',
    'Noweb',
    'NPL-1.0',
    'NPL-1.1',
    'NPOSL-3.0',
    'NRL',
    'NTP',
    'Nunit',
    'OCCT-PL',
    'OCLC-2.0',
    'ODbL-1.0',
    'OFL-1.0',
    'OFL-1.1',
    'OGTSL',
    'OLDAP-1.1',
    'OLDAP-1.2',
    'OLDAP-1.3',
    'OLDAP-1.4',
    'OLDAP-2.0.1',
    'OLDAP-2.0',
    'OLDAP-2.1',
    'OLDAP-2.2.1',
    'OLDAP-2.2.2',
    'OLDAP-2.2',
    'OLDAP-2.3',
    'OLDAP-2.4',
    'OLDAP-2.5',
    'OLDAP-2.6',
    'OLDAP-2.7',
    'OLDAP-2.8',
    'OML',
    'OpenSSL',
    'OPL-1.0',
    'OSET-PL-2.1',
    'OSL-1.0',
    'OSL-1.1',
    'OSL-2.0',
    'OSL-2.1',
    'OSL-3.0',
    'PDDL-1.0',
    'PHP-3.01',
    'PHP-3.0',
    'Plexus',
    'PostgreSQL',
    'psfrag',
    'psutils',
    'Python-2.0',
    'Qhull',
    'QPL-1.0',
    'Rdisc',
    'RHeCos-1.1',
    'RPL-1.1',
    'RPL-1.5',
    'RPSL-1.0',
    'RSA-MD',
    'RSCPL',
    'Ruby',
    'Saxpath',
    'SAX-PD',
    'SCEA',
    'Sendmail',
    'SGI-B-1.0',
    'SGI-B-1.1',
    'SGI-B-2.0',
    'SimPL-2.0',
    'SISSL-1.2',
    'SISSL',
    'Sleepycat',
    'SMLNJ',
    'SMPPL',
    'SNIA',
    'Spencer-86',
    'Spencer-94',
    'Spencer-99',
    'SPL-1.0',
    'StandardML-NJ',
    'SugarCRM-1.1.3',
    'SWL',
    'TCL',
    'TCP-wrappers',
    'TMate',
    'TORQUE-1.1',
    'TOSL',
    'Unicode-DFS-2015',
    'Unicode-DFS-2016',
    'Unicode-TOU',
    'Unlicense',
    'UPL-1.0',
    'Vim',
    'VOSTROM',
    'VSL-1.0',
    'W3C-19980720',
    'W3C-20150513',
    'W3C',
    'Watcom-1.0',
    'Wsuipa',
    'WTFPL',
    'wxWindows',
    'X11',
    'Xerox',
    'XFree86-1.1',
    'xinetd',
    'Xnet',
    'xpp',
    'XSkat',
    'YPL-1.0',
    'YPL-1.1',
    'Zed',
    'Zend-2.0',
    'Zimbra-1.3',
    'Zimbra-1.4',
    'zlib-acknowledgement',
    'Zlib',
    'ZPL-1.1',
    'ZPL-2.0',
    'ZPL-2.1',

    # Exceptions
    '389-exception',
    'Autoconf-exception-2.0',
    'Autoconf-exception-3.0',
    'Bison-exception-2.2',
    'Bootloader-exception',
    'Classpath-exception-2.0',
    'CLISP-exception-2.0',
    'DigiRule-FOSS-exception',
    'eCos-exception-2.0',
    'Fawkes-Runtime-exception',
    'FLTK-exception',
    'Font-exception-2.0',
    'freertos-exception-2.0',
    'GCC-exception-2.0',
    'GCC-exception-3.1',
    'gnu-javamail-exception',
    'i2p-gpl-java-exception',
    'Libtool-exception',
    'Linux-syscall-note',
    'LZMA-exception',
    'mif-exception',
    'Nokia-Qt-exception-1.1',
    'OCCT-exception-1.0',
    'openvpn-openssl-exception',
    'Qwt-exception-1.0',
    'u-boot-exception-2.0',
    'WxWindows-exception-3.1',
]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2017  Free Software Foundation Europe e.V.
#
# This file is part of reuse, available from its original location:
# <https://git.fsfe.org/reuse/reuse/>.
#
# reuse is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# reuse is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# reuse.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Global fixtures and configuration."""

# pylint: disable=redefined-outer-name

import logging
import os
import shutil
import subprocess
from collections import namedtuple
from io import StringIO
from pathlib import Path
from typing import Dict, Optional, Tuple

import jinja2
import pytest
from reuse import ReuseInfo
from reuse._util import GIT_EXE, GIT_METHOD, setup_logging

CWD = Path.cwd()

TESTS_DIRECTORY = Path(__file__).parent.resolve()
RESOURCES_DIRECTORY = TESTS_DIRECTORY / 'resources'
CODE_FILES_DIRECTORY = RESOURCES_DIRECTORY / 'code_files'

# Some licenses to test against
LICENSES = [
    'CC0-1.0',
    'GPL-3.0',
    'GPL-3.0+',
    'GPL-3.0-only',
    'GPL-3.0-or-later',
    '(GPL-2.0-only OR BSD-3-Clause)',
]

GIT_METHODS = [None]
if GIT_EXE:
    GIT_METHODS.append('git')
if GIT_METHOD == 'pygit2':
    GIT_METHODS.append('pygit2')


NameAndLicense = namedtuple(
    'NameAndLicense',
    ['name', 'reuse_info'],
)


def pytest_configure(config):
    """Called after command line options have been parsed and all plugins and
    initial conftest files been loaded.
    """
    if config.getoption('--capture') == 'no':
        setup_logging(level=logging.DEBUG)


def pytest_runtest_setup(item):
    """Called before running a test."""
    # pylint: disable=unused-argument
    # Make sure to restore CWD
    os.chdir(str(CWD))


def render_code_files() -> Dict[NameAndLicense, str]:
    """Compile all code files with Jinja2 and return a dictionary with the
    file as key, and rendered text as value.
    """
    loader = jinja2.FileSystemLoader(str(CODE_FILES_DIRECTORY))
    environment = jinja2.Environment(loader=loader)

    result = dict()

    for file_ in CODE_FILES_DIRECTORY.iterdir():
        if not file_.is_file():
            continue
        if file_.suffix == '.license':
            continue

        template = environment.get_template(file_.name)

        for license in LICENSES:
            context = {
                'license': license,
            }

            # Put some related information in a struct-like object.
            name_and_license = NameAndLicense(
                '{}___{}'.format(license, file_.name),
                ReuseInfo(
                    (context['license'],),
                    ('Copyright (C) 2017  Free Software Foundation Europe '
                     'e.V.',)))

            result[name_and_license] = template.render(context)

    return result


COMPILED_CODE_FILES = render_code_files()


@pytest.fixture(autouse=True, params=GIT_METHODS)
def git_method(request, monkeypatch) -> Optional[str]:
    """Run the test with multiple methods of git usage."""
    monkeypatch.setattr('reuse.GIT_METHOD', request.param)
    monkeypatch.setattr('reuse._util.GIT_METHOD', request.param)
    yield request.param


@pytest.fixture()
def tiny_repository(tmpdir_factory) -> Path:
    """Create a tiny temporary fake repository."""
    directory = Path(str(tmpdir_factory.mktemp('tiny')))
    src = directory / 'src'
    debian_dir = directory / 'debian'
    licenses_dir = directory / 'LICENSES'
    src.mkdir()
    debian_dir.mkdir()
    licenses_dir.mkdir()

    text = """
    # Copyright (C) 2017  Free Software Foundation Europe e.V.
    #
    # SPDX-License-Identifier: GPL-3.0+
    """
    (src / 'code.py').write_text(text)
    (src / 'no_license.py').touch()

    shutil.copy(
        str(RESOURCES_DIRECTORY / 'debian/copyright'),
        str(debian_dir / 'copyright'))

    # Fake text
    (licenses_dir / 'GPL-3.0.txt').write_text('GPL-3.0')

    os.chdir(str(directory))
    return directory


@pytest.fixture()
def empty_directory(tmpdir_factory) -> Path:
    """Create a temporary empty directory."""
    directory = Path(str(tmpdir_factory.mktemp('empty_directory')))

    os.chdir(str(directory))
    return directory


@pytest.fixture()
def fake_repository(tmpdir_factory) -> Path:
    """Create a temporary fake repository."""
    directory = Path(str(tmpdir_factory.mktemp('fake_repository')))
    src = directory / 'src'
    src.mkdir()
    debian_dir = directory / 'debian'
    debian_dir.mkdir()
    licenses_dir = directory / 'LICENSES'
    licenses_dir.mkdir()
    doc = directory / 'doc'
    doc.mkdir()

    rendered_texts = COMPILED_CODE_FILES

    for name_and_license, text in rendered_texts.items():
        (src / name_and_license.name).write_text(text)

    # debian/copyright
    shutil.copy(
        str(RESOURCES_DIRECTORY / 'debian/copyright'),
        str(debian_dir / 'copyright'))
    (doc / 'index.rst').touch()

    (directory / 'README.md').write_text(
        """
        # Copyright (C) 2017  Free Software Foundation Europe e.V.
        #
        # SPDX-License-Identifier: CC0-1.0
        # License-Filename: LICENSES/CC0-1.0
        """)

    # TODO: Write full licence texts
    licenses = ['CC0-1.0', 'GPL-3.0', 'GPL-2.0', 'BSD-3-Clause']
    for license in licenses:
        (licenses_dir / '{}.txt'.format(license)).touch()

    os.chdir(str(directory))
    return directory


@pytest.fixture()
def git_repository(fake_repository: Path, git_method: Optional[str]) -> Path:
    """Create a git repository with ignored files."""
    if git_method is None:
        pytest.skip('cannot run this test without git')

    subprocess.run(['git', 'init', str(fake_repository)])

    gitignore = "*.pyc\nbuild"
    (fake_repository / '.gitignore').write_text(gitignore)

    for file_ in (fake_repository / 'src').iterdir():
        if file_.suffix == '.py':
            file_.with_suffix('.pyc').touch()

    build_dir = fake_repository / 'build'
    build_dir.mkdir()
    (build_dir / 'hello.py').touch()

    os.chdir(str(fake_repository))
    return fake_repository


@pytest.fixture(params=COMPILED_CODE_FILES.items())
def file_with_license_comments(request) -> StringIO:
    """Provide a code file that has REUSE license information in its header
    comments.

    The code file is a fake file (StringIO).  It contains additional attributes
    for the test to read.
    """
    key, value = request.param
    result = StringIO(value)
    result.name = key.name
    result.reuse_info = key.reuse_info
    yield result


@pytest.fixture(params=COMPILED_CODE_FILES.items(), scope='session')
def empty_file_with_license_file(
        request, tmpdir_factory) -> Tuple[Path, ReuseInfo]:
    """Create a temporary directory that contains two files:  The code file and
    the license file.
    """
    directory = Path(str(tmpdir_factory.mktemp('empty_file_with_license')))

    key, value = request.param

    (directory / '{}.license'.format(key.name)).write_text(value)
    (directory / key.name).touch()

    os.chdir(str(directory))
    return (directory, key.reuse_info)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018  Carmen Bianca Bakker
#
# This file is part of reuse, available from its original location:
# <https://git.fsfe.org/reuse/reuse/>.
#
# reuse is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# reuse is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# reuse.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Formatting functions primarily for the CLI."""

from textwrap import indent, fill

WIDTH = 78
INDENT = 2


def fill_paragraph(text, width=WIDTH, indent_width=0):
    """Wrap a single paragraph."""
    return indent(
        fill(text.strip(), width=width - indent_width), indent_width * ' ')


def fill_all(text, width=WIDTH, indent_width=0):
    """Wrap all paragraphs."""
    return '\n\n'.join(
        fill_paragraph(paragraph, width=width, indent_width=indent_width)
        for paragraph in split_into_paragraphs(text))


def split_into_paragraphs(text):
    """Yield all paragraphs in a text.  A paragraph is a piece of text
    surrounded by empty lines.
    """
    lines = text.splitlines()
    paragraph = ''

    for line in lines:
        if not line:
            if paragraph:
                yield paragraph
                paragraph = ''
            else:
                continue
        else:
            if paragraph:
                padding = ' '
            else:
                padding = ''
            paragraph = '{}{}{}'.format(paragraph, padding, line)
    if paragraph:
        yield paragraph



# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2018  Free Software Foundation Europe e.V.
#
# This file is part of reuse, available from its original location:
# <https://git.fsfe.org/reuse/reuse/>.
#
# reuse is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# reuse is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# reuse.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Tests for reuse."""

import os
import shutil
from itertools import zip_longest
from pathlib import Path
from textwrap import dedent

import pytest

import reuse
from reuse import _util

# pylint: disable=invalid-name
git = pytest.mark.skipif(
    not _util.GIT_EXE,
    reason='requires git')


# Set of licenses used in tests.  This is a bit of duplication from conftest.
USED_LICENSES = set(['CC0-1.0', 'GPL-3.0', 'GPL-2.0', 'BSD-3-Clause'])


def _reuse_info_equal(first, second) -> bool:
    """Compare two ReuseInfo objects.

    This is necessary because (,) != [].
    """
    for left, right in zip_longest(first, second):
        if tuple(left) != tuple(right):
            return False
    return True


def test_extract_license_from_file(file_with_license_comments):
    """Test whether you can correctly extract reuse information from a code
    file's comments.
    """
    result = reuse.extract_reuse_info(
        file_with_license_comments.getvalue())
    assert _reuse_info_equal(result, file_with_license_comments.reuse_info)


def test_extract_no_license_info():
    """Given a file without reuse information, return an empty ReuseInfo
    object.
    """
    result = reuse.extract_reuse_info('')
    assert _reuse_info_equal(result, reuse.ReuseInfo([], []))


def test_extract_expression():
    """Parse various expressions."""
    expressions = [
        'GPL-3.0+',
        'GPL-3.0 AND CC0-1.0',
        'nonsense',
    ]
    for expression in expressions:
        result = reuse.extract_reuse_info(
            'SPDX-License-'
            'Identifier: {}'.format(expression))
        assert result.spdx_expressions == {expression}


def test_project_not_a_directory(empty_directory):
    """Cannot create a Project without a valid directory."""
    (empty_directory / 'foo.py').touch()
    with pytest.raises(NotADirectoryError):
        reuse.Project(empty_directory / 'foo.py')


def test_reuse_info_of_file_does_not_exist(fake_repository):
    """Raise FileNotFoundError when asking for the reuse info of a file that
    does not exist.
    """
    project = reuse.Project(fake_repository)
    with pytest.raises(FileNotFoundError):
        project.reuse_info_of(fake_repository / 'does_not_exist')


def test_reuse_info_of_directory(empty_directory):
    """Raise IsADirectoryError when calling reuse_info_of on a directory."""
    (empty_directory / 'src').mkdir()

    project = reuse.Project(empty_directory)
    with pytest.raises(IsADirectoryError):
        project.reuse_info_of(empty_directory / 'src')


def test_reuse_info_of_unlicensed_file(fake_repository):
    """Return an empty ReuseInfo object when asking for the reuse information
    of a file that has no reuse information.
    """
    (fake_repository / 'foo.py').touch()
    project = reuse.Project(fake_repository)
    assert not any(project.reuse_info_of('foo.py'))


def test_reuse_info_of_only_copyright(fake_repository):
    """A file contains only a copyright line.  Test whether it correctly picks
    up on that.
    """
    (fake_repository / 'foo.py').write_text('Copyright (C) 2017  Mary Sue')
    project = reuse.Project(fake_repository)
    reuse_info = project.reuse_info_of('foo.py')
    assert not any(reuse_info.spdx_expressions)
    assert len(reuse_info.copyright_lines) == 1
    assert reuse_info.copyright_lines.pop() == 'Copyright (C) 2017  Mary Sue'


def test_reuse_info_of_only_copyright_also_covered_by_debian(fake_repository):
    """A file contains only a copyright line, but debian/copyright also has
    information on this file.  Use both.
    """
    (fake_repository / 'doc/foo.py').write_text('Copyright in file')
    project = reuse.Project(fake_repository)
    reuse_info = project.reuse_info_of('doc/foo.py')
    assert any(reuse_info.spdx_expressions)
    assert len(reuse_info.copyright_lines) == 2
    assert 'Copyright in file' in reuse_info.copyright_lines
    assert '2017 Mary Sue' in reuse_info.copyright_lines


def test_reuse_info_of_also_covered_by_debian(fake_repository):
    """A file contains all reuse information, but debian/copyright also
    provides information on this file.  Use both.
    """
    (fake_repository / 'doc/foo.py').write_text(
        dedent("""
            SPDX-License-Identifier: GPL-3.0
            Copyright in file"""))
    project = reuse.Project(fake_repository)
    reuse_info = project.reuse_info_of('doc/foo.py')
    for thing in reuse_info:
        assert len(thing) == 2
    assert 'GPL-3.0' in reuse_info.spdx_expressions
    assert 'CC0-1.0' in reuse_info.spdx_expressions
    assert 'Copyright in file' in reuse_info.copyright_lines
    assert '2017 Mary Sue' in reuse_info.copyright_lines


def test_reuse_info_of_no_duplicates(empty_directory):
    """A file contains the same lines twice.  The ReuseInfo only contains those
    lines once.
    """
    spdx_line = 'SPDX-License-Identifier: GPL-3.0+\n'
    copyright_line = 'Copyright (C) 2017  Free Software Foundation Europe\n'
    text = spdx_line + copyright_line

    (empty_directory / 'foo.py').write_text(text * 2)
    project = reuse.Project(empty_directory)
    reuse_info = project.reuse_info_of('foo.py')
    for thing in reuse_info:
        assert len(thing) == 1
    assert 'GPL-3.0+' in reuse_info.spdx_expressions
    assert copyright_line.strip() in reuse_info.copyright_lines


def test_error_in_debian_copyright(fake_repository):
    """If there is an error in debian/copyright, just ignore its existence."""
    (fake_repository / 'debian/copyright').write_text('invalid')
    project = reuse.Project(fake_repository)
    assert not any(project.reuse_info_of('doc/index.rst'))


def test_all_files(empty_directory):
    """Given a directory with some files, yield all files."""
    (empty_directory / 'foo').touch()
    (empty_directory / 'bar').touch()

    project = reuse.Project(empty_directory)
    assert {file_.name for file_ in project.all_files()} == {'foo', 'bar'}


def test_all_files_on_single_file(empty_directory):
    """When a file is given as parameter instead of a directory, just yield the
    file.
    """
    (empty_directory / 'foo').touch()

    project = reuse.Project(empty_directory)
    result = list(project.all_files(empty_directory / 'foo'))

    assert len(result) == 1
    assert result[0].name == 'foo'


def test_license_file_detected(empty_file_with_license_file):
    """Test whether---when given a file and a license file---the license file
    is detected and read.
    """
    directory = empty_file_with_license_file[0]
    reuse_info = empty_file_with_license_file[1]

    project = reuse.Project(directory)

    all_files = list(project.all_files(directory))
    assert len(all_files) == 1

    result = project.reuse_info_of(all_files[0])
    assert _reuse_info_equal(result, reuse_info)


def test_all_licensed(fake_repository):
    """Given a repository where all files are licensed, check if
    Project.lint yields nothing.
    """
    project = reuse.Project(fake_repository)

    assert not list(project.lint())


def test_all_licensed_from_different_cwd(fake_repository):
    """Same as the other test, but try a different CWD."""
    os.chdir('/')
    project = reuse.Project(fake_repository)

    assert not list(project.lint())


def test_empty_directory_is_licensed(empty_directory):
    """An empty directory is licensed."""
    project = reuse.Project(empty_directory)

    assert not list(project.lint())


def test_all_licensed_no_debian_copyright(fake_repository):
    """The exact same as test_all_licensed, but now without
    debian/copyright.
    """
    shutil.rmtree(str(fake_repository / 'debian'))
    os.remove(str(fake_repository / 'doc/index.rst'))

    project = reuse.Project(fake_repository)

    assert not list(project.lint())


def test_one_unlicensed(fake_repository):
    """Given a repository where one file is not licensed, check if
    Project.unlicensed yields that file.
    """
    (fake_repository / 'foo.py').touch()

    project = reuse.Project(fake_repository)

    assert list(project.lint()) == [fake_repository / 'foo.py']


def test_all_licensed_but_unknown_license(fake_repository):
    """If a file points to a license that does not exist, it is unlicensed."""
    (fake_repository / 'foo.py').write_text(
        'SPDX-License-Identifier: LicenseRef-foo')

    project = reuse.Project(fake_repository)

    assert list(project.lint()) == [fake_repository / 'foo.py']


def test_all_licensed_but_error_in_spdx_expression(fake_repository):
    """If a file contains an SPDX expression that cannot be parsed, it is
    unlicensed.
    """
    (fake_repository / 'foo.py').write_text(
        'SPDX-License-Identifier: this is an invalid expression')

    project = reuse.Project(fake_repository)

    assert list(project.lint()) == [fake_repository / 'foo.py']


def test_lint_only_copyright(empty_directory):
    """If a file has only copyright information associated with it, it is
    unlicensed.
    """
    (empty_directory / 'foo.py').write_text(
        'Copyright (C) 2017  Mary Sue')

    project = reuse.Project(empty_directory)

    assert project.lint_file(empty_directory / 'foo.py')
    assert not project.lint_file(
        empty_directory / 'foo.py',
        spdx_mandatory=False)


def test_lint_only_spdx(empty_directory):
    """If a file has only SPDX information associated with it, it is
    unlicensed.
    """
    (empty_directory / 'foo.py').write_text(
        'SPDX-License-Identifier: MIT')
    (empty_directory / 'COPYING').write_text(
        'Valid-License-Identifier: MIT')

    project = reuse.Project(empty_directory)

    assert project.lint_file(empty_directory / 'foo.py')
    assert not project.lint_file(
        empty_directory / 'foo.py',
        copyright_mandatory=False)


def test_lint_not_a_file(fake_repository):
    """lint_file raises an OSError when called on a non-file."""
    project = reuse.Project(fake_repository)
    with pytest.raises(OSError):
        project.lint_file(fake_repository / 'src')


def test_lint_license_not_found(empty_directory):
    """If a license is used that doesn't appear in Project.licenses, lint
    complains.
    """
    (empty_directory / 'foo.py').write_text(
        'Copyright (C) 2017  Mary Sue\n'
        'SPDX-License-Identifier: MIT')
    project = reuse.Project(empty_directory)

    assert any(project.lint())


def test_licenses_from_filenames(fake_repository):
    """Given a repository, extract the license identifiers from the
    filenames.
    """
    project = reuse.Project(fake_repository)

    assert set(project.licenses.keys()) == USED_LICENSES
    assert set(
        map(str, project.licenses.values())) == \
        set(['LICENSES/{}.txt'.format(spdx) for spdx in USED_LICENSES])


def test_licenses_licenseref_from_filename(empty_directory):
    """Extract SPDX identifier from filename if it begins with 'LicenseRef-'"""
    (empty_directory / 'LICENSES').mkdir()
    (empty_directory / 'LICENSES/LicenseRef-hello.txt').touch()

    project = reuse.Project(empty_directory)

    assert set(project.licenses.keys()) == {'LicenseRef-hello'}


def test_licenses_no_file_extension(fake_repository):
    """Given a license file with no extension, correctly identify it."""
    (fake_repository / 'LICENSES/GPL-3.0.txt').rename(
        fake_repository / 'LICENSES/GPL-3.0')
    project = reuse.Project(fake_repository)

    assert set(project.licenses.keys()) == USED_LICENSES


@pytest.mark.parametrize(
    'license_file',
    ['COPYING', 'COPYING.md', 'LICENSE', 'LICENCE', 'COPYRIGHT',
     'LICENSES/MIT.txt', 'LICENSES/subdir/MIT.txt'])
def test_license_from_tag(empty_directory, license_file):
    """Extract license identifiers from the license file itself."""
    (empty_directory / license_file).parent.mkdir(parents=True, exist_ok=True)
    (empty_directory / license_file).write_text(
        'Valid-License-Identifier: MIT')

    project = reuse.Project(empty_directory)
    assert list(project.licenses.keys()) == ['MIT']
    assert list(project.licenses.values()) == [Path(license_file)]


def test_licenses_priority(empty_directory):
    """.license files are prioritised over the files themselves."""
    (empty_directory / 'COPYING').write_text('Valid-License-Identifier: MIT')
    (empty_directory / 'COPYING.license').write_text(
        'Valid-License-Identifier: GPL-3.0')

    project = reuse.Project(empty_directory)
    assert list(project.licenses.keys()) == ['GPL-3.0']


def test_licenses_from_different_cwd(empty_directory):
    """If the CWD is different, still provide correct licenses."""
    os.chdir('/')
    (empty_directory / 'COPYING').write_text('Valid-License-Identifier: MIT')

    project = reuse.Project(empty_directory)
    assert list(project.licenses.keys()) == ['MIT']


def test_licenses_multiple_in_file(empty_directory):
    """If there are multiple licenses in a file, return them all."""
    (empty_directory / 'COPYING').write_text(
        'Valid-License-Identifier: GPL-3.0\n'
        'Valid-License-Identifier: GPL-3.0+')

    project = reuse.Project(empty_directory)
    assert set(project.licenses.keys()) == {'GPL-3.0', 'GPL-3.0+'}


def test_licenses_unknown(empty_directory):
    """If a license has no known license identifier, create one for it."""
    (empty_directory / 'COPYING').touch()
    (empty_directory / 'LICENSES').mkdir()
    (empty_directory / 'LICENSES/NOT-IN-SPDX-LIST.txt').touch()

    project = reuse.Project(empty_directory)
    assert set(project.licenses.keys()) == {
        'LicenseRef-Unknown0', 'LicenseRef-Unknown1'}


def test_licenses_conflict(empty_directory):
    """If two license files both resolve to the same identifier, raise a
    RuntimeError.
    """
    (empty_directory / 'COPYING').write_text(
        'Valid-License-Identifier: MIT')
    (empty_directory / 'LICENSE').write_text(
        'Valid-License-Identifier: MIT')

    project = reuse.Project(empty_directory)
    with pytest.raises(RuntimeError):
        assert project.licenses


@git
def test_unlicensed_but_ignored_by_git(git_repository):
    """Given a Git repository where some files are unlicensed---but ignored by
    git---check if Project.unlicensed yields nothing.
    """
    project = reuse.Project(git_repository)

    assert not list(project.lint())


@git
def test_find_root_in_git_repo(git_repository):
    """When using reuse from a child directory in a Git repo, always find the
    root directory.
    """
    os.chdir('src')
    result = _util.find_root()

    assert Path(result).absolute().resolve() == git_repository


def test_encoding():
    """Given a source code file, correctly detect its encoding and read it."""
    tests_directory = Path(__file__).parent.resolve()
    encoding_directory = tests_directory / 'resources/encoding'
    project = reuse.Project(encoding_directory)

    for path in encoding_directory.iterdir():
        reuse_info = project.reuse_info_of(path)
        assert reuse_info.copyright_lines.pop() == 'Copyright © 2017  Liberté'



# -*- coding: utf-8 -*-
#
# Copyright (C) 2017  Free Software Foundation Europe e.V.
#
# This file is part of reuse, available from its original location:
# <https://git.fsfe.org/reuse/reuse/>.
#
# reuse is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# reuse is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# reuse.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Misc. utilities for reuse."""

import logging
import os
import shutil
import subprocess
from pathlib import Path
from typing import BinaryIO, List, Optional, Union

import chardet


GIT_EXE = shutil.which('git')


try:
    from pygit2 import GitError, discover_repository
    GIT_METHOD = 'pygit2'
except ImportError:  # pragma: no cover
    if GIT_EXE:
        GIT_METHOD = 'git'
    else:
        GIT_METHOD = None


_logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


PathLike = Union[Path, str]  # pylint: disable=invalid-name


def setup_logging(level: int = logging.WARNING) -> None:
    """Configure logging for reuse."""
    # library_logger is the root logger for reuse.  We configure logging solely
    # for reuse, not for any other libraries.
    library_logger = logging.getLogger('reuse')
    library_logger.setLevel(level)

    if not library_logger.hasHandlers():
        handler = logging.StreamHandler()
        formatter = logging.Formatter('%(name)s - %(levelname)s - %(message)s')
        handler.setFormatter(formatter)
        library_logger.addHandler(handler)


def execute_command(
        command: List[str],
        logger: logging.Logger,
        **kwargs) -> subprocess.CompletedProcess:
    """Run the given command with subprocess.run.  Forward kwargs.  Silence
    output into a pipe unless kwargs override it.
    """
    logger.debug('running %s', ' '.join(command))

    stdout = kwargs.get('stdout', subprocess.PIPE)
    stderr = kwargs.get('stderr', None)

    return subprocess.run(
        command,
        stdout=stdout,
        stderr=stderr,
        **kwargs)


def find_root() -> Optional[Path]:
    """Try to find the root of the project from $PWD.  If none is found, return
    None.
    """
    cwd = Path.cwd()
    if in_git_repo(cwd):
        if GIT_METHOD == 'pygit2':
            repo = discover_repository(str(cwd))
            return Path(repo).parent
        elif GIT_METHOD == 'git':
            command = [GIT_EXE, 'rev-parse', '--show-toplevel']
            result = execute_command(command, _logger, cwd=str(cwd))

            if not result.returncode:
                path = result.stdout.decode('utf-8')[:-1]
                return Path(os.path.relpath(path, str(cwd)))
    return None


def in_git_repo(cwd: PathLike = None) -> bool:
    """Is *cwd* inside of a git repository?

    Always return False if git is not installed.
    """
    if cwd is None:
        cwd = Path.cwd()

    if GIT_METHOD == 'pygit2':
        try:
            discover_repository(str(cwd))
            return True
        except KeyError:
            return False
    elif GIT_METHOD == 'git':
        command = [GIT_EXE, 'status']
        result = execute_command(command, _logger, cwd=str(cwd))

        return not result.returncode

    return False


def _is_binary_string(bytes_string: bytes) -> bool:
    """Given a bytes object, does it look like a binary string or a text
    string?

    Behaviour is based on file(1).
    """
    textchars = bytearray(
        {7, 8, 9, 10, 12, 13, 27} | set(range(0x20, 0x100)) - {0x7f})
    return bool(bytes_string.translate(None, textchars))


def decoded_text_from_binary(binary_file: BinaryIO, size: int = None) -> str:
    """Given a binary file object, detect its encoding and return its contents
    as a decoded string.  Do not throw any errors if the encoding contains
    errors:  Just replace the false characters.

    If *size* is specified, only read so many bytes.
    """
    rawdata = binary_file.read(size)
    if _is_binary_string(rawdata):
        raise UnicodeError('cannot decode binary data')
    result = chardet.detect(rawdata)
    encoding = result.get('encoding')
    if encoding is None:
        encoding = 'utf-8'
    try:
        return rawdata.decode(encoding, errors='replace')
    # Handle unknown encodings.
    except LookupError as error:
        raise UnicodeError('could not decode {}'.format(encoding)) from error



# -*- coding: utf-8 -*-
#
# Copyright (C) 2017  Free Software Foundation Europe e.V.
#
# This file is part of reuse, available from its original location:
# <https://git.fsfe.org/reuse/reuse/>.
#
# reuse is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# reuse is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# reuse.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Tests for the CLI for reuse."""

from io import StringIO
from itertools import zip_longest
from textwrap import dedent

import pytest
from reuse import __version__, _main, _util

# pylint: disable=invalid-name
git = pytest.mark.skipif(
    not _util.GIT_EXE,
    reason='requires git')


def test_lint_none(fake_repository):
    """Given a repository in which every file is licensed, return an exit code
    of 0 and print nothing.
    """
    out = StringIO()
    result = _main.main(['lint', str(fake_repository)], out)

    assert result == 0
    assert not out.getvalue()


@git
def test_lint_gitignore(git_repository):
    """Given a repository with files ignored by Git, skip over those files."""
    out = StringIO()
    result = _main.main(['lint', str(git_repository)], out)

    assert result == 0
    assert not out.getvalue()


def test_lint_ignore_debian(fake_repository):
    """When debian/copyright is ignored, non-compliant files are found."""
    out = StringIO()
    result = _main.main(['--ignore-debian', 'lint', str(fake_repository)], out)

    output_lines = out.getvalue().splitlines()
    assert len(output_lines) == 1
    assert 'index.rst' in output_lines[0]
    assert result


def test_lint_twice_path(fake_repository):
    """When providing the same path twice, only output those files once."""
    (fake_repository / 'foo.py').touch()
    (fake_repository / 'bar.py').touch()
    out = StringIO()
    result = _main.main(['lint'] + [str(fake_repository / 'foo.py')] * 2, out)

    output_lines = out.getvalue().splitlines()
    assert len(output_lines) == 1
    assert 'foo.py' in output_lines[0]
    assert result


@pytest.mark.xfail(reason='Order of files is not reliable')
def test_compile(tiny_repository):
    """A correct bill of materials is generated."""
    # result = runner.invoke(
    #     _main.cli,
    #     ['compile'])
    result = None

    expected = dedent("""\
        SPDXVersion: SPDX-2.1
        DataLicense: CC0-1.0
        SPDXID: SPDXRef-DOCUMENT
        DocumentName: {dirname}
        DocumentNamespace: http://spdx.org/spdxdocs/spdx-v2.1-04c223f0-4415-47fd-9860-7074a07f753e
        Creator: Person: Anonymous ()
        Creator: Organization: Anonymous ()
        Creator: Tool: reuse-{version}
        Created: 2017-11-08T11:07:30Z
        CreatorComment: <text>This document was created automatically using available reuse information consistent with the REUSE Initiative.</text>
        Relationship: SPDXRef-DOCUMENT describes SPDXRef-8008eeb8d2000e5aa6eaa51b1cdc944d726e1107
        Relationship: SPDXRef-DOCUMENT describes SPDXRef-bb5656f1b5e8283a8e930c54afd9a8bfebe7a548

        FileName: ./src/code.py
        SPDXID: SPDXRef-8008eeb8d2000e5aa6eaa51b1cdc944d726e1107
        FileChecksum: SHA1: fb17c6dd60b8b8c35128c8a14905e1ef328b1534
        LicenseConcluded: NOASSERTION
        LicenseInfoInFile: GPL-3.0+
        FileCopyrightText: <text>Copyright (C) 2017  Free Software Foundation Europe e.V.</text>

        FileName: ./src/no_license.py
        SPDXID: SPDXRef-bb5656f1b5e8283a8e930c54afd9a8bfebe7a548
        FileChecksum: SHA1: da39a3ee5e6b4b0d3255bfef95601890afd80709
        LicenseConcluded: NOASSERTION
        LicenseInfoInFile: CC0-1.0
        FileCopyrightText: <text>2017 Mary Sue</text>""".format(
            dirname=tiny_repository.name,
            version=__version__))

    for result_line, expected_line in zip_longest(
            result.output.splitlines(),
            expected.splitlines()):
        # Just ignore these
        if 'DocumentNamespace' in result_line:
            continue
        if 'Created' in result_line:
            continue

        assert result_line == expected_line



# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2018  Free Software Foundation Europe e.V.
# Copyright (C) 2018  Carmen Bianca Bakker
#
# This file is part of reuse, available from its original location:
# <https://git.fsfe.org/reuse/reuse/>.
#
# reuse is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# reuse is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# reuse.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""reuse is a tool for compliance with the REUSE Initiative recommendations."""

# pylint: disable=ungrouped-imports,too-many-arguments

import contextlib
import datetime
import gettext
import glob
import hashlib
import logging
import os
import re
import sys
from gettext import gettext as _
from pathlib import Path
from typing import (BinaryIO, Dict, Iterator, List, NamedTuple, Optional, Set,
                    Union)
from uuid import uuid4

import pkg_resources
from debian.copyright import Copyright, NotMachineReadableError

from ._util import (GIT_EXE, GIT_METHOD, PathLike, decoded_text_from_binary,
                    execute_command, in_git_repo)
from .licenses import LICENSES

try:
    from pygit2 import Repository, GitError
except ImportError:  # pragma: no cover
    pass

_LOCALE_DIRS = [
    # sys.prefix is usually /usr, but can also be the root of the virtualenv.
    sys.prefix + '/share/locale',
    # Relevant for `pip install --user` installations.
    str(Path.home()) + '/.local/share/locale',
    # This somehow works for egg installations.
    pkg_resources.resource_filename(
        pkg_resources.Requirement.parse('fsfe-reuse'),
        'share/locale'),
]

for dir in _LOCALE_DIRS:
    # 'eo' is only used here because I am certain that this translation exists.
    if (Path(dir) / 'eo/LC_MESSAGES/reuse.mo').exists():
        gettext.bindtextdomain('reuse', dir)
        gettext.textdomain('reuse')
        break

__author__ = 'Carmen Bianca Bakker'
__email__ = 'carmenbianca@fsfe.org'
__license__ = 'GPLv3+'
__version__ = '0.3.1'

_logger = logging.getLogger(__name__)  # pylint: disable=invalid-name

# Amount of bytes that we assume will be big enough to contain the entire
# comment header (including SPDX tags), so that we don't need to read the
# entire file.
_HEADER_BYTES = 4096

_END_PATTERN = r'(?:\*/)*(?:-->)*$'
_LICENSE_PATTERN = re.compile(
    r'SPDX-Licen[cs]e-Identifier: (.*?)' + _END_PATTERN,
    re.MULTILINE)
_COPYRIGHT_PATTERN = re.compile(
    r'(Copyright .*?)' + _END_PATTERN,
    re.MULTILINE)
_VALID_LICENSE_PATTERN = re.compile(
    r'Valid-Licen[cs]e-Identifier: (.*?)' + _END_PATTERN,
    re.MULTILINE)

_IGNORE_DIR_PATTERNS = [
    re.compile(r'^\.git$'),
    re.compile(r'^\.svn$'),
    re.compile(r'^LICEN[CS]ES$'),
]

_IGNORE_FILE_PATTERNS = [
    re.compile(r'^LICEN[CS]E'),
    re.compile(r'^COPYING'),
    re.compile(r'^copyright$'),
    re.compile(r'.*\.license$'),
    re.compile(r'.*\.spdx$'),
    re.compile(r'^\.gitignore$'),
]

#: Simple structure for holding REUSE information.
ReuseInfo = NamedTuple(
    'ReuseInfo',
    [
        ('spdx_expressions', Set[str]),
        ('copyright_lines', Set[str])
    ])


class ReuseException(Exception):
    """Base exception."""


class IdentifierNotFound(ReuseException):
    """Could not find SPDX identifier for license file."""


def _checksum(file_object: BinaryIO, hash_function) -> str:
    """Return checksum of *file_object*."""
    result = hash_function()
    for chunk in iter(
            lambda: file_object.read(128 * result.block_size), b''):
        result.update(chunk)
    return result.hexdigest()


def _copyright_from_debian(
        path: PathLike,
        copyright: Copyright) -> ReuseInfo:
    """Find the reuse information of *path* in the Debian copyright object.
    """
    result = copyright.find_files_paragraph(str(path))

    if result is None:
        return ReuseInfo(set(), set())

    return ReuseInfo(
        set([result.license.synopsis]),
        set(map(str.strip, result.copyright.splitlines())))


def _identifiers_from_expression(expression: str) -> List[str]:
    """Given an SPDX expression, return a list of the identifiers within.

    >>> _identifiers_from_expression('MIT AND (GPL-3.0+ OR CC0-1.0)')
    ['MIT', 'GPL-3.0+', 'CC0-1.0']
    """
    # All substrings that need to be removed for just the identifiers to
    # remain.
    to_replace = ['(', ')']
    for substring in to_replace:
        expression = expression.replace(substring, '')

    boolean_words = ['OR', 'AND']
    for word in boolean_words:
        expression = re.sub(
            r'\s{}\s'.format(word), ' ', expression, flags=re.IGNORECASE)

    return expression.split()


def _strip_gpl_extension(identifier: str) -> List[str]:
    """Strip '+', '-only' and '-or-later' from *identifier*."""
    to_remove = ['+', '-only', '-or-later']
    for string in to_remove:
        if identifier.endswith(string):
            identifier = identifier.replace(string, '')
    return identifier


def _determine_license_path(path: PathLike) -> Path:
    """Given a path FILE, return FILE.license if it exists, otherwise return
    FILE.
    """
    path = Path(path)
    license_path = Path('{}.license'.format(path))
    if not license_path.exists():
        license_path = path
    return license_path


def extract_reuse_info(text: str) -> ReuseInfo:
    """Extract reuse information from comments in a string."""
    license_matches = set(map(str.strip, _LICENSE_PATTERN.findall(text)))
    copyright_matches = set(map(str.strip, _COPYRIGHT_PATTERN.findall(text)))

    return ReuseInfo(
        license_matches,
        copyright_matches)


def extract_valid_license(text: str) -> Set[str]:
    """Extract SPDX identifier from a string."""
    return set(map(str.strip, _VALID_LICENSE_PATTERN.findall(text)))


class Project:  # pylint: disable=unused-variable
    """Simple object that holds the project's root, which is necessary for many
    interactions.
    """

    def __init__(self, root: PathLike):
        self._root = Path(root)
        if not self._root.is_dir():
            raise NotADirectoryError('{} is no valid path'.format(self._root))

        self._git_repo = None
        if GIT_METHOD == 'pygit2':
            with contextlib.suppress(GitError):
                self._git_repo = Repository(str(self._root))
        elif GIT_METHOD == 'git':
            self._git_repo = in_git_repo(self._root)
        else:
            _logger.warning(_('could not find Git'))
        self._license_files = None
        # Use '0' as None, because None is a valid value...
        self._copyright_val = 0

    def all_files(self, directory: PathLike = None) -> Iterator[Path]:
        """Yield all files in *directory* and its subdirectories.

        The files that are not yielded are:

        - Files ignored by VCS (e.g., see .gitignore)

        - Files/directories matching IGNORE_*_PATTERNS.

        If *directory* is a file, yield it if it is not ignored.
        """
        if directory is None:
            directory = self.root
        directory = Path(directory)

        if directory.is_file() and not self._is_path_ignored(directory):
            # Translators: %s is a directory path.  This is inside a loop that
            # yields all files that will be checked for REUSE information
            # (i.e., some files are ignored, and thus not yielded).
            _logger.debug(_('yielding %s'), directory)
            yield directory

        for root, dirs, files in os.walk(str(directory)):
            root = Path(root)
            # Translators: %s is a directory path.
            _logger.debug(_('currently walking in %s'), root)

            # Don't walk ignored directories
            for dir_ in list(dirs):
                if self._is_path_ignored(root / dir_):
                    # Translators: %s is a path.
                    _logger.debug(_('ignoring %s'), root / dir_)
                    dirs.remove(dir_)

            # Filter files.
            for file_ in files:
                if self._is_path_ignored(root / file_):
                    _logger.debug(_('ignoring %s'), root / file_)
                    continue

                _logger.debug(_('yielding %s'), file_)
                yield root / file_

    def reuse_info_of(
            self,
            path: PathLike,
            ignore_debian: bool = False) -> ReuseInfo:
        """Get the reuse information of *path*.

        This function will return any reuse information that it can find, both
        from within the file and from the debian/copyright file.  If none is
        found, an empty ReuseInfo object is returned.
        """
        path = _determine_license_path(path)
        # Translators: %s is a path.
        _logger.debug(_('searching %s for reuse information'), path)

        spdx_expressions = set()
        copyright_lines = set()

        with path.open('rb') as fp:
            try:
                file_result = extract_reuse_info(
                    decoded_text_from_binary(fp, size=_HEADER_BYTES))
                spdx_expressions = spdx_expressions.union(
                    file_result.spdx_expressions)
                copyright_lines = copyright_lines.union(
                    file_result.copyright_lines)
            except UnicodeError:
                # Translators: %s is a path.
                _logger.info(_('%s could not be decoded'), path)

        # Search the debian/copyright file for copyright information.
        if not ignore_debian and self._copyright:
            debian_result = _copyright_from_debian(
                self._relative_from_root(path),
                self._copyright)
            if any(debian_result):
                # Translators: %s is a path.
                _logger.info(_('%s covered by debian/copyright'), path)
                spdx_expressions = spdx_expressions.union(
                    debian_result.spdx_expressions)
                copyright_lines = copyright_lines.union(
                    debian_result.copyright_lines)

        return ReuseInfo(spdx_expressions, copyright_lines)

    def lint(
            self,
            path: PathLike = None,
            spdx_mandatory: bool = True,
            copyright_mandatory: bool = True,
            ignore_debian: bool = False,
            ignore_missing: bool = False) -> Iterator[Path]:
        """Yield all files under *path* that are not compliant with the REUSE
        recommendations.

        If *path* is not specified, it becomes root.

        The modifiers are explained in :meth:`~Project.lint_file`.
        """
        if path is None:
            path = self.root
        for file_ in self.all_files(path):
            try:
                if self.lint_file(
                        file_,
                        spdx_mandatory=spdx_mandatory,
                        copyright_mandatory=copyright_mandatory,
                        ignore_debian=ignore_debian,
                        ignore_missing=ignore_missing):
                    yield file_
            except OSError:
                # Translators: %s is a path.
                _logger.error(_('Could not read %s'), file_)
                yield file_

    def lint_file(
            self,
            path: PathLike,
            spdx_mandatory: bool = True,
            copyright_mandatory: bool = True,
            ignore_debian: bool = False,
            ignore_missing: bool = False) -> int:
        """
        :param path: A path to a file.  If it is not a file, raise an OSError.
        :param spdx_mandatory: The file must have an SPDX expression in its
            reuse information.
        :param copyright_mandatory: The file must have a copyright line in its
            reuse information.
        :param ignore_debian: copyright/debian will not be checked for reuse
            information.
        :param ignore_missing: Declared licences in SPDX expression that could
            not be found in :attr:`~Project.licenses` will not affect the
            linter.

        Check whether *path* complies with the REUSE recommendations.  If it
        does, return 0.  If it does not, return a non-zero integer.  In a
        future version, the non-zero integer may be a bit mask.

        If both *spdx_mandatory* and *copyright_mandatory* are false, this
        function more or less becomes useless.  Common sense is advised.
        """
        path = Path(path)
        if not path.is_file():
            raise OSError('{} is not a file'.format(path))

        reuse_info = self.reuse_info_of(
            path,
            ignore_debian=ignore_debian)

        if spdx_mandatory and not any(reuse_info.spdx_expressions):
            return 1
        if copyright_mandatory and not any(reuse_info.copyright_lines):
            return 1

        if not ignore_missing:
            for expression in reuse_info.spdx_expressions:
                wrong_identifier = self._contains_invalid_identifier(
                    expression)
                if wrong_identifier:
                    _logger.warning(_(
                        '{path} is licensed under {identifier}, but its '
                        'license file could not be found').format(
                            path=path, identifier=wrong_identifier))

                    return 1

        return 0

    def bill_of_materials(
            self,
            out=sys.stdout,
            ignore_debian: bool = False) -> None:
        """Generate a bill of materials from the project.  The bill of
        materials is written to *out*.

        See https://spdx.org/specifications.
        """
        # Write mandatory tags
        out.write('SPDXVersion: SPDX-2.1\n')
        out.write('DataLicense: CC0-1.0\n')
        out.write('SPDXID: SPDXRef-DOCUMENT\n')

        out.write('DocumentName: {}\n'.format(self.root.resolve().name))
        # TODO: Generate UUID from git revision maybe
        # TODO: Fix the URL
        out.write(
            'DocumentNamespace: '
            'http://spdx.org/spdxdocs/spdx-v2.1-{}\n'.format(uuid4()))

        # Author
        # TODO: Fix Person and Organization
        out.write('Creator: Person: Anonymous ()\n')
        out.write('Creator: Organization: Anonymous ()\n')
        out.write('Creator: Tool: reuse-{}\n'.format(__version__))

        now = datetime.datetime.utcnow()
        now = now.replace(microsecond=0)
        out.write('Created: {}Z\n'.format(now.isoformat()))
        out.write(
            'CreatorComment: <text>This document was created automatically '
            'using available reuse information consistent with the '
            'REUSE Initiative.</text>\n')

        all_files = sorted(self.all_files())

        # List all DESCRIBES relationships.  This involves some code
        # duplication in determining the relative path to the file and its
        # hash.
        for file_ in all_files:
            relative = self._relative_from_root(file_)
            ref = hashlib.sha1(str(relative).encode('utf-8')).hexdigest()
            out.write(
                'Relationship: SPDXRef-DOCUMENT describes '
                'SPDXRef-{}\n'.format(ref))

        # File information
        for file_ in all_files:
            out.write('\n')
            self._file_information(file_, out, ignore_debian=ignore_debian)

        # Licenses
        for license, path in self.licenses.items():
            if license.startswith('LicenseRef-'):
                out.write('\n')
                out.write('LicenseID: {}\n'.format(license))
                out.write('LicenseName: NOASSERTION\n')

                with (self.root / path).open() as fp:
                    out.write(
                        'ExtractedText: <text>{}</text>\n'.format(fp.read()))

    @property
    def licenses(self) -> Dict[str, Path]:
        """Return a dictionary of all licenses in the project, with their SPDX
        identifiers as names and paths as values.

        If no name could be found for a license file, name it
        "LicenseRef-Unknown0" and count upwards for every other unknown file.
        """
        if self._license_files is not None:
            return self._license_files

        unknown_counter = 0
        license_files = dict()

        patterns = [
            'LICENSE*', 'LICENCE*', 'COPYING*', 'COPYRIGHT*', 'LICENCES/**',
            'LICENSES/**']
        for pattern in patterns:
            pattern = str(self.root.resolve() / pattern)
            for path in glob.iglob(pattern, recursive=True):
                # For some reason, LICENSES/** is resolved even though it
                # doesn't exist.  I have no idea why.  Deal with that here.
                if not Path(path).exists() or Path(path).is_dir():
                    continue
                if Path(path).suffix == '.license':
                    continue

                path = _determine_license_path(path)
                path = self._relative_from_root(path)
                # Translators: %s is a path.
                _logger.debug(_('searching %s for license tags'), path)

                try:
                    identifiers = self._identifiers_of_license(path)
                except IdentifierNotFound:
                    identifier = 'LicenseRef-Unknown{}'.format(unknown_counter)
                    identifiers = [identifier]
                    unknown_counter += 1
                    _logger.warning(_(
                        'Could not resolve SPDX identifier of {path}, '
                        'resolving to {identifier}').format(
                            path=path, identifier=identifier))

                for identifier in identifiers:
                    if identifier in license_files:
                        _logger.critical(_(
                            '{identifier} is the SPDX identifier of both '
                            '{path} and {other_path}').format(
                                identifier=identifier, path=path,
                                other_path=license_files[identifier]))
                        raise RuntimeError(
                            'Multiple licenses resolve to {}'.format(
                                identifier))
                    license_files[identifier] = path

        self._license_files = license_files
        return self._license_files

    @property
    def root(self) -> Path:
        """Path to the root of the project."""
        return self._root

    @property
    def _copyright(self) -> Optional[Copyright]:
        if self._copyright_val == 0:
            copyright_path = self.root / 'debian' / 'copyright'
            try:
                with copyright_path.open() as fp:
                    self._copyright_val = Copyright(fp)
            except (IOError, OSError):
                _logger.debug(
                    _('no debian/copyright file, or could not read it'))
            except NotMachineReadableError:
                _logger.exception(_('debian/copyright has syntax errors'))

            # This check is a bit redundant, but otherwise I'd have to repeat
            # this line under each exception.
            if not self._copyright_val:
                self._copyright_val = None
        return self._copyright_val

    def _contains_invalid_identifier(
            self,
            expression: str) -> Union[bool, str]:
        """Is the expression an invalid SPDX expression?  i.e., does any
        identifier refer to a file that does not exist in Project.licenses?

        Return the faulty identifier.

        If all identifiers are valid, return False.
        """
        identifiers = _identifiers_from_expression(expression)

        for identifier in identifiers:
            if (
                    identifier not in self.licenses
                    and _strip_gpl_extension(identifier) not in self.licenses):
                return identifier
        return False

    def _identifiers_of_license(self, path: PathLike) -> List[str]:
        """Figure out the SPDX identifier(s) of a license given its path.

        The order of precedence is:

        - A .license file containing the `Valid-License-Identifier` tag.

        - A `Valid-License-Identifier` tag within the license file itself.

        - The name of the file (minus extension) if:

          - The name is an SPDX license.

          - The name starts with 'LicenseRef-'.
        """
        path = _determine_license_path(path)

        with (self.root / path).open('rb') as fp:
            result = extract_valid_license(
                decoded_text_from_binary(fp, size=_HEADER_BYTES))
            if any(result):
                return result

        for name in (path.stem, path.name):
            if name in LICENSES:
                return [name]
        if path.stem.startswith('LicenseRef-'):
            return [path.stem]

        raise IdentifierNotFound(
            'Could not find SPDX identifier for {}'.format(path))

    def _ignored_by_git(self, path: PathLike) -> bool:
        """Is *path* covered by the ignore mechanism of git?

        Always return False if git is not installed.
        """
        path = self._relative_from_root(path)

        if GIT_METHOD == 'pygit2':
            return self._git_repo.path_is_ignored(str(path))
        elif GIT_METHOD == 'git':
            command = [GIT_EXE, 'check-ignore', str(path)]

            result = execute_command(command, _logger, cwd=str(self.root))
            return not result.returncode
        return False

    def _ignored_by_vcs(self, path: PathLike) -> bool:
        """Is *path* covered by the ignore mechanism of the VCS (e.g.,
        .gitignore)?
        """
        if self._git_repo:
            return self._ignored_by_git(path)
        return False

    def _is_path_ignored(self, path: PathLike) -> bool:
        """Is *path* ignored by some mechanism?"""
        path = Path(path)

        if path.is_file():
            for pattern in _IGNORE_FILE_PATTERNS:
                if pattern.match(path.name):
                    return True
        elif path.is_dir():
            for pattern in _IGNORE_DIR_PATTERNS:
                if pattern.match(path.name):
                    return True

        if self._ignored_by_vcs(path):
            return True

        return False

    def _relative_from_root(self, path: PathLike) -> Path:
        """If the project root is /tmp/project, and *path* is
        /tmp/project/src/file, then return src/file.
        """
        return Path(os.path.relpath(str(path), start=str(self.root)))

    def _file_information(
            self,
            path: PathLike,
            out=sys.stdout,
            ignore_debian: bool = False) -> None:
        """Create SPDX File Information for *path*."""
        relative = self._relative_from_root(path)
        encoded = str(relative).encode('utf-8')
        out.write('FileName: ./{}\n'.format(relative))
        out.write('SPDXID: SPDXRef-{}\n'.format(
            hashlib.sha1(encoded).hexdigest()))

        with path.open('rb') as fp:
            out.write(
                'FileChecksum: SHA1: {}\n'.format(
                    _checksum(fp, hashlib.sha1)))

        # IMPORTANT: Make no assertion about concluded license.  This tool
        # cannot, with full certainty, determine the license of a file.
        out.write('LicenseConcluded: NOASSERTION\n')

        reuse_info = self.reuse_info_of(path, ignore_debian=ignore_debian)

        for spdx in sorted(reuse_info.spdx_expressions):
            out.write('LicenseInfoInFile: {}\n'.format(spdx))

        if reuse_info.copyright_lines:
            for line in sorted(reuse_info.copyright_lines):
                out.write('FileCopyrightText: <text>{}</text>\n'.format(line))
        else:
            out.write('FileCopyrightText: NONE\n')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017  Free Software Foundation Europe e.V.
# Copyright (C) 2018  Carmen Bianca Bakker
#
# This file is part of reuse, available from its original location:
# <https://git.fsfe.org/reuse/reuse/>.
#
# reuse is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# reuse is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# reuse.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import os
import glob
import subprocess
from pathlib import Path

from setuptools import setup

requirements = [
    'chardet',
    'python-debian',
]
git_extras = []

if not os.environ.get('REUSE_DEV'):
    git_extras.append('pygit2')

test_requirements = [
    'pytest',
    'jinja2',
]


def readme_rst():
    try:
        command = ['pandoc', 'README.md', '-t', 'rst']
        result = subprocess.run(command, stdout=subprocess.PIPE)
        return result.stdout.decode('utf-8')
    except FileNotFoundError:
        if os.environ.get('RST_ERROR'):
            raise
        return open('README.md').read()


def mo_files():
    paths = glob.glob('po/**/**/reuse.mo')
    return [
        ('share/locale/{}/LC_MESSAGES'.format(Path(path).parts[1]), [path])
        for path in paths]


if __name__ == '__main__':
    setup(
        name='fsfe-reuse',
        version='0.3.1',
        url='https://git.fsfe.org/reuse/reuse',
        license='GPL-3.0-or-later',

        author='Carmen Bianca Bakker',
        author_email='carmenbianca@fsfe.org',

        description='reuse is a tool for compliance with the REUSE Initiative '
            'recommendations.',
        long_description=readme_rst(),

        package_dir={
            '': 'src'
        },
        packages=[
            'reuse',
        ],

        data_files=mo_files(),

        entry_points={
            'console_scripts': [
                'reuse = reuse._main:main',
            ],
        },

        install_requires=requirements,
        tests_require=test_requirements,
        extras_require={
            'git': git_extras,
        },

        classifiers=[
            'Development Status :: 3 - Alpha',
            'Intended Audience :: Developers',
            'License :: OSI Approved :: '
            'GNU General Public License v3 or later (GPLv3+)',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
        ],
    )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2017  Free Software Foundation Europe e.V.
# Copyright (C) 2018  Carmen Bianca Bakker
#
# This file is part of reuse, available from its original location:
# <https://git.fsfe.org/reuse/reuse/>.
#
# reuse is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# reuse is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# reuse.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Entry functions for reuse."""

import argparse
import importlib
import logging
import sys
from gettext import gettext as _
from pathlib import Path
from pipes import quote
from typing import List

from ._format import INDENT, fill_all, fill_paragraph
from ._util import GIT_METHOD, find_root, setup_logging

# Import __init__.py.  I don't know how to do this cleanly
reuse = importlib.import_module('..', __name__)  # pylint: disable=invalid-name

_logger = logging.getLogger(__name__)  # pylint: disable=invalid-name

_DESCRIPTION_LINES = [
    _('reuse  Copyright (C) 2017-2018  Free Software Foundation Europe e.V.'),

    _(
        'reuse is a tool for compliance with the REUSE Initiative '
        'recommendations.  See <https://reuse.software/> for more '
        'information.'),

    _(
        # Translators: Find an (un)official translation of the GPL for this
        # bit.
        'reuse is free software: you can redistribute it and/or modify it '
        'under the terms of the GNU General Public License as published by '
        'the Free Software Foundation, either version 3 of the License, or '
        '(at your option) any later version.\n'
        '\n'
        'reuse is distributed in the hope that it will be useful, but WITHOUT '
        'ANY WARRANTY; without even the implied warranty of MERCHANTABILITY '
        'or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public '
        'License for more details.\n'
        '\n'
        'You should have received a copy of the GNU General Public License '
        'along with reuse.  If not, see <http://www.gnu.org/licenses/>.'),

    _("Support the FSFE's work:")]

_INDENTED_LINE = _(
    'Donations are critical to our strength and autonomy.  They enable us to '
    'continue working for Free Software wherever necessary.  Please consider '
    'making a donation at <https://fsfe.org/donate/>.')

_DESCRIPTION_TEXT = (
    fill_all('\n\n'.join(_DESCRIPTION_LINES))
    + '\n\n'
    + fill_paragraph(_INDENTED_LINE, indent_width=INDENT))


_EPILOG_TEXT = ''
_PYGIT2_WARN = '\n\n'.join([
    _('IMPORTANT:'),

    fill_paragraph(
        _(
            'You do not have pygit2 installed.  reuse will slow down '
            'significantly because of this.  For better performance, please '
            "install your distribution's version of pygit2."),
        indent_width=INDENT)])
if not GIT_METHOD == 'pygit2':
    _EPILOG_TEXT = _EPILOG_TEXT + '\n\n' + _PYGIT2_WARN


def _create_project() -> reuse.Project:
    """Create a project object.  Try to find the project root from $PWD,
    otherwise treat $PWD as root.
    """
    root = find_root()
    if root is None:
        root = Path.cwd()
    return reuse.Project(root)


def compile(args, out=sys.stdout):
    """Print the project's bill of materials."""
    project = _create_project()
    if args.output:
        out = args.output
        if not out.name.endswith('.spdx'):
            # Translators: %s is a file name.
            _logger.warning(_('%s does not end with .spdx'), out.name)
    project.bill_of_materials(
        out,
        ignore_debian=args.ignore_debian)

    return 0


def license(args, out=sys.stdout):
    """Print the SPDX expressions of each provided file."""
    project = _create_project()
    first = True

    for path in args.paths:
        try:
            reuse_info = project.reuse_info_of(
                path,
                ignore_debian=args.ignore_debian)
        except IsADirectoryError:
            _logger.error(_('%s is a directory'), path)
            continue
        except IOError:
            # Translators: %s is a file.
            _logger.error(_('could not read %s'), path)
            continue

        if not first:
            out.write('\n')

        out.write(quote(str(path)))
        out.write('\n')

        if any(reuse_info.spdx_expressions):
            out.write(', '.join(map(quote, reuse_info.spdx_expressions)))
            out.write('\n')
        else:
            out.write(_('none\n'))

        first = False

    return 0


def lint(args, out=sys.stdout):
    """List all non-compliant files."""
    counter = 0
    found = set()

    project = _create_project()
    paths = args.paths
    if not paths:
        paths = [project.root]

    for path in paths:
        for file_ in project.lint(
                path,
                spdx_mandatory=args.spdx_mandatory,
                copyright_mandatory=args.copyright_mandatory,
                ignore_debian=args.ignore_debian,
                ignore_missing=args.ignore_missing):
            output = quote(str(file_))
            if output not in found:
                out.write(output)
                out.write('\n')
                found.add(output)
                counter += 1

    return counter


def parser() -> argparse.ArgumentParser:
    """Create the parser and return it."""
    # pylint: disable=redefined-outer-name
    parser = argparse.ArgumentParser(
        'reuse', formatter_class=argparse.RawDescriptionHelpFormatter,
        description=(_DESCRIPTION_TEXT), epilog=_EPILOG_TEXT)
    parser.add_argument(
        '--debug', action='store_true', help=_('enable debug statements'))
    parser.add_argument(
        '--ignore-debian', action='store_true',
        help=_('do not use debian/copyright to extract reuse information'))
    parser.add_argument(
        '--version', action='store_true',
        help=_("show program's version number and exit"))
    parser.set_defaults(func=lambda x, y: parser.print_help())

    subparsers = parser.add_subparsers()

    compile_parser = subparsers.add_parser(
        'compile',
        help=_("print the project's bill of materials"))
    compile_parser.add_argument(
        '--output', '-o', action='store', type=argparse.FileType('w'))
    compile_parser.set_defaults(func=compile)

    lint_parser = subparsers.add_parser(
        'lint', formatter_class=argparse.RawDescriptionHelpFormatter,
        help=_('list all non-compliant files'),
        description=fill_all(_(
            'List all non-compliant files.\n'
            '\n'
            'A file is non-compliant when:\n'
            '\n'
            '- It has no copyright information.\n'
            '\n'
            '- It has no license (declared as SPDX expression).\n'
            '\n'
            '- Its license could not be found.\n'
            '\n'
            'This prints only the paths of the files that do not comply, each '
            'file on a separate line.\n'
            '\n'
            'Error and warning messages are output to STDERR.')))
    lint_parser.add_argument(
        'paths', action='store', nargs='*')
    lint_parser.add_argument(
        '--spdx-mandatory', action='store_true', default=True,
        help=_('SPDX expressions are mandatory for compliance'))
    lint_parser.add_argument(
        '--copyright-mandatory', action='store_true', default=True,
        help=_('copyright notices are mandatory for compliance'))
    lint_parser.add_argument(
        '--ignore-missing', action='store_true',
        help=_('ignore missing licenses'))
    lint_parser.set_defaults(func=lint)

    license_parser = subparsers.add_parser(
        'license',
        help=_('print the SPDX expressions of each provided file'))
    license_parser.add_argument(
        'paths', action='store', nargs='*')
    license_parser.set_defaults(func=license)

    return parser


def main(args: List[str] = None, out=sys.stdout) -> None:
    """Main entry function."""
    if args is None:
        args = sys.argv[1:]

    main_parser = parser()
    parsed_args = main_parser.parse_args(args)

    setup_logging(
        level=logging.DEBUG if parsed_args.debug else logging.WARNING)

    if parsed_args.version:
        out.write(_('reuse, version {}\n').format(reuse.__version__))
        return 0
    return parsed_args.func(parsed_args, out)
